@@ -16,6 +16,9 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
+
+#include <config.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -34,6 +37,7 @@
 static int status_fd, passphrase_fd;
 static GString *status_commands;
 static GHashTable *userid_hints;
+static GHashTable *bad_passphrases;
 static char *passphrase_keyid = NULL;
 
 static void
@@ -64,31 +68,34 @@ error_dialog (const char *fmt, ...)
 
 /* send the passphrase for the given key ID to the passphrase FD */
 static void
-handle_get_passphrase (const char *keyid, const char *userid)
+handle_get_passphrase (const char *keyid, const char *userid,
+		       gboolean bad_passphrase)
 {
   GnomeKeyringAttributeList *attributes;
   GnomeKeyringResult keyring_result;
   GList *found_items = NULL;
   char *keyring = NULL, *passphrase;
-  gboolean save_password = FALSE;
+  gboolean keyring_available, save_password = FALSE;
   guint32 item_id;
 
-  attributes = gnome_keyring_attribute_list_new ();
-  gnome_keyring_attribute_list_append_string (attributes,
-					      "gnome-gpg-keyid",
-					      keyid);
-  keyring_result = gnome_keyring_find_items_sync (
-      GNOME_KEYRING_ITEM_GENERIC_SECRET,
-      attributes, &found_items);
-  gnome_keyring_attribute_list_free (attributes);
+  keyring_available = gnome_keyring_is_available();
 
-  if (keyring_result != GNOME_KEYRING_RESULT_OK
-      && keyring_result != GNOME_KEYRING_RESULT_DENIED) {
-    error_dialog (_("Couldn't search keyring (code %d)"),
-		  (int) keyring_result);
-    close (passphrase_fd);
-    gtk_main_quit ();
-    return;
+  if (keyring_available) {
+    attributes = gnome_keyring_attribute_list_new ();
+    gnome_keyring_attribute_list_append_string (
+        attributes, GNOME_GPG_KEYID_ATTR, keyid);
+    keyring_result = gnome_keyring_find_items_sync (
+        GNOME_KEYRING_ITEM_GENERIC_SECRET, attributes, &found_items);
+    gnome_keyring_attribute_list_free (attributes);
+
+    if (keyring_result != GNOME_KEYRING_RESULT_OK
+	&& keyring_result != GNOME_KEYRING_RESULT_DENIED) {
+      error_dialog (_("Couldn't search keyring (code %d)"),
+		    (int) keyring_result);
+      close (passphrase_fd);
+      gtk_main_quit ();
+      return;
+    }
   }
 
   if (found_items == NULL) {
@@ -105,9 +112,19 @@ handle_get_passphrase (const char *keyid, const char *userid)
       short_keyid = keyid;
 
     if (userid) {
-      message = g_strdup_printf(_("Please enter the passphrase for user \"%s\" (%s)"), userid, short_keyid);
+      message = g_strdup_printf (
+          _("Please enter the passphrase for \"%s\" (%s)"),
+          userid, short_keyid);
     } else {
-      message = g_strdup_printf(_("Please enter the passphrase for key %s"), short_keyid);
+      message = g_strdup_printf(
+          _("Please enter the passphrase for key %s"), short_keyid);
+    }
+    /* if the previous attempt gave a bad passphrase, try again. */
+    if (bad_passphrase) {
+      char *tmp = g_strconcat(_("Invalid passphrase; please try again.\n\n"),
+			      message, NULL);
+      g_free(message);
+      message = tmp;
     }
     dialog = GNOME_PASSWORD_DIALOG (gnome_password_dialog_new (
         _("GNU Privacy Guard passphrase"),
@@ -119,7 +136,7 @@ handle_get_passphrase (const char *keyid, const char *userid)
     gnome_password_dialog_set_show_userpass_buttons (dialog, FALSE);
     gnome_password_dialog_set_show_username (dialog, FALSE);
     gnome_password_dialog_set_show_domain (dialog, FALSE);
-    gnome_password_dialog_set_show_remember (dialog, TRUE);
+    gnome_password_dialog_set_show_remember (dialog, keyring_available);
 
     zenity_util_show_dialog (GTK_WIDGET (dialog));
     if (!gnome_password_dialog_run_and_block (dialog)) {
@@ -154,16 +171,16 @@ handle_get_passphrase (const char *keyid, const char *userid)
   write (passphrase_fd, passphrase, strlen (passphrase));
   write (passphrase_fd, "\n", 1);
 
-  if (save_password) {
+  if (keyring_available && save_password) {
     gchar *nice_name = g_strdup_printf (_("GNU Privacy Guard passphrase for key ID: %s"),
 					keyid);
     attributes = gnome_keyring_attribute_list_new ();
     gnome_keyring_attribute_list_append_string (attributes,
-						"gnome-gpg-keyid",
+						GNOME_GPG_KEYID_ATTR,
 						keyid);
     if (userid) {
       gnome_keyring_attribute_list_append_string (attributes,
-						  "gnome-gpg-userid",
+						  GNOME_GPG_USERID_ATTR,
 						  userid);
     }
     keyring_result = gnome_keyring_item_create_sync (keyring,
@@ -191,10 +208,14 @@ handle_bad_passphrase (const char *keyid)
   GnomeKeyringAttributeList *attributes;
   GnomeKeyringResult keyring_result;
   GList *found_items = NULL, *tmp;
+
+  /* if the keyring is not available, we can't invalidate the item */
+  if (!gnome_keyring_is_available())
+    return;
   
   attributes = gnome_keyring_attribute_list_new ();
   gnome_keyring_attribute_list_append_string (attributes,
-					      "gnome-gpg-keyid",
+					      GNOME_GPG_KEYID_ATTR,
 					      keyid);
   keyring_result = gnome_keyring_find_items_sync (
       GNOME_KEYRING_ITEM_GENERIC_SECRET,
@@ -251,10 +272,15 @@ handle_command (const char *command)
     }
   } else if (!strcmp(command, "GET_HIDDEN passphrase.enter")) {
     char *userid;
+    gboolean bad_passphrase;
 
     /* get the userid hint for this key, if any */
     userid = g_hash_table_lookup(userid_hints, passphrase_keyid);
-    handle_get_passphrase(passphrase_keyid, userid);
+    /* was a bad passphrase entered? */
+    bad_passphrase = GPOINTER_TO_INT(g_hash_table_lookup(
+        bad_passphrases, passphrase_keyid));
+    g_hash_table_remove(bad_passphrases, passphrase_keyid);
+    handle_get_passphrase(passphrase_keyid, userid, bad_passphrase);
   } else if (!strncmp(command, "GET", 3)) {
     fprintf(stderr, "unhandled GET issued: %s\n", command);
     close(passphrase_fd);
@@ -268,6 +294,8 @@ handle_command (const char *command)
     } else {
       keyid = g_strdup(command);
     }
+    g_hash_table_insert (bad_passphrases, g_strdup (keyid),
+			 GINT_TO_POINTER(TRUE));
     handle_bad_passphrase(keyid);
     g_free(keyid);
   } else {
@@ -323,13 +351,6 @@ main (int argc, char **argv)
   GIOChannel *channel;
   char *short_argv[2];
 
-#if 0
-  if (!gnome_keyring_is_available ()) {
-    fprintf (stderr, "gnome-keyring is not available\n");
-    exit (execvp ("gpg", argv));
-  }
-#endif
-
   if (pipe (passphrase_fds) < 0) {
     g_error ("Couldn't pipe: %s", g_strerror (errno));
     exit (1);
@@ -368,7 +389,7 @@ main (int argc, char **argv)
     }
     g_ptr_array_add (args, NULL);
 
-    if (execvp ("gpg", (char **) g_ptr_array_free (args, FALSE)) < 0) {
+    if (execv (GPG, (char **) g_ptr_array_free (args, FALSE)) < 0) {
       fprintf (stderr, "Couldn't exec gpg: %s", g_strerror (errno));
     }
     return 1;
@@ -399,6 +420,11 @@ main (int argc, char **argv)
 					(GDestroyNotify)g_free,
 					(GDestroyNotify)g_free);
 
+  /* keys for which a bad passphrase has been given */
+  bad_passphrases = g_hash_table_new_full (g_str_hash, g_str_equal,
+					   (GDestroyNotify)g_free,
+					   (GDestroyNotify)NULL);
+
   /* buffer of status commands from gpg */
   status_commands = g_string_new (NULL);
 
